/* 
 * ----- LCD Abstraction Layer: ----- 
 */

#ifdef LCD_ETPP
// Eagle Tree Power Panel  http://www.eagletreesystems.com/MicroPower/9.htm
// The Power Panel is a physically small & light weight, 2x16, transflective (readable in sunlight) LCD that interfaces via i2c bus (two wire).
//   White wire  to Ground
//   Red wire    to +5V
//   Yellow wire to A4
//   Brown wire  to A5
// ETPP code contributed by Danal Estes

// *********************
// I2C general functions
// *********************
  #define I2C_SPEED 100000L     // 100kHz normal mode, this value must be used for a genuine WMP
  #define INTERNAL_I2C_PULLUPS  // Comment this if your hardware supplies i2c pullups outside the Amtel chip
  #define I2C_PULLUPS_ENABLE         PORTC |= 1<<4; PORTC |= 1<<5;   // PIN A4&A5 (SDA&SCL)
  #define I2C_PULLUPS_DISABLE        PORTC &= ~(1<<4); PORTC &= ~(1<<5);
  
  // Mask prescaler bits : only 5 bits of TWSR defines the status of each I2C request
  #define TW_STATUS_MASK	(1<<TWS7) | (1<<TWS6) | (1<<TWS5) | (1<<TWS4) | (1<<TWS3)
  #define TW_STATUS             (TWSR & TW_STATUS_MASK)
  
  void i2c_init(void) {
    #if defined(INTERNAL_I2C_PULLUPS)
      I2C_PULLUPS_ENABLE
    #else
      I2C_PULLUPS_DISABLE
    #endif
    TWSR = 0;        // no prescaler => prescaler = 1
    TWBR = ((16000000L / I2C_SPEED) - 16) / 2; // change the I2C clock rate
    TWCR = 1<<TWEN;  // enable twi module, no interrupt
  }
  
  void i2c_rep_start(uint8_t address) {
    TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN)  ;//| (1<<TWSTO); // send REAPEAT START condition
    waitTransmissionI2C(); // wait until transmission completed
    checkStatusI2C(); // check value of TWI Status Register
    TWDR = address; // send device address
    TWCR = (1<<TWINT) | (1<<TWEN);
    waitTransmissionI2C(); // wail until transmission completed
    checkStatusI2C(); // check value of TWI Status Register
  }
  
  void i2c_write(uint8_t data ) {	
    TWDR = data; // send data to the previously addressed device
    TWCR = (1<<TWINT) | (1<<TWEN);
    waitTransmissionI2C(); // wait until transmission completed
    checkStatusI2C(); // check value of TWI Status Register
  }
  
  uint8_t i2c_readAck() {
    TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA);
    waitTransmissionI2C();
    return TWDR;
  }
  
  uint8_t i2c_readNak(void) {
    TWCR = (1<<TWINT) | (1<<TWEN);
    waitTransmissionI2C();
    return TWDR;
  }
  
  void waitTransmissionI2C() {
    uint8_t count = 255;
    while (count-->0 && !(TWCR & (1<<TWINT)) );
    if (count<2) { //we are in a blocking state => we don't insist
      TWCR = 0;  //and we force a reset on TWINT register
    }
  }
  
  void checkStatusI2C() {
    uint8_t tstat = TW_STATUS;
    if ( tstat == 0xF8) { //TW_NO_INFO : this I2C error status indicates a wrong I2C communication.
      exit (1);  //Fatal. 
    } 
  }

// *********************
// I2C Eagle Tree Power Panel primitives
// *********************
  void i2c_ETPP_init () {
    i2c_rep_start(0x76+0);      // I2C write direction
    i2c_write(0x00);            // ETPP command register
    i2c_write(0x24);            // Function Set 001D0MSL D : data length for parallel interface only; M: 0 = 1x32 , 1 = 2x16; S: 0 = 1:18 multiplex drive mode, 1�32 or 2�16 character display, 1 = 1:9 multiplex drive mode, 1�16 character display; H: 0 = basic instruction set plus standard instruction set, 1 = basic instruction set plus extended instruction set
    i2c_write(0x0C);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
    i2c_write(0x06);            // Cursor Move  000001IS I : 0 = DDRAM or CGRAM address decrements by 1, cursor moves to the left, 1 = DDRAM or CGRAM address increments by 1, cursor moves to the right; S : 0 = display does not shift,  1 = display does shifts
    lcdClear();         
  }

  void i2c_ETPP_send_cmd (byte c) {
    i2c_rep_start(0x76+0);      // I2C write direction
    i2c_write(0x00);            // ETPP command register
    i2c_write(c);
  }

  void i2c_ETPP_send_char (char c) {
    if (c > 0x0f) c |=  0x80;   // ETPP uses character set "R", which has A->z mapped same as ascii + high bit; don't mess with custom chars. 
    i2c_rep_start(0x76+0);      // I2C write direction
    i2c_write(0x40);            // ETPP data register
    i2c_write(c);
  }

  void i2c_ETPP_send_string (const char *s) {
    i2c_rep_start(0x76+0);      // I2C write direction
    i2c_write(0x40);            // ETPP data register
    while (*s) {byte c = *s++; if (c > 0x0f) c |=  0x80; i2c_write(c);}
  }

  void i2c_ETPP_set_cursor (byte col, byte row) {
    byte addr = col + row * 0x40;
    i2c_ETPP_send_cmd(0x80 | addr);     // High bit is "Set DDRAM" command, remaing bits are addr.  0x40 = first char of second line when in 2-line display (aka MUX 1:9 mode)
  }

  void i2c_ETPP_create_char (byte idx, uint8_t* array) {
    i2c_ETPP_send_cmd(0x80);             // CGRAM and DDRAM share an address register, but you can't set certain bits with the CGRAM address command.   Use DDRAM address command to be sure high order address bits are zero. 
    i2c_ETPP_send_cmd(0x40 | byte(idx * 8));   // Set CGRAM address 
    i2c_rep_start(0x76+0);               // I2C write direction
    i2c_write(0x40);                     // ETPP data register
    for (byte i = 0 ; i<8 ; i++) {i2c_write(*array); array++;}
  }

// *********************
// LCD functions similar to LiquidCrystal lib functions, as adapted for ETPP
// *********************
  void lcdClear() {
    i2c_ETPP_send_cmd(0x01);                              //Clear display command, which does NOT clear an Eagle Tree because character set "R" has a '>' at 0x20
    for (byte i = 0; i<80; i++) i2c_ETPP_send_char(' ');  // Blanks for all 80 bytes of RAM in the controller, not just the 2x16 display
  }
  
  void lcdBlink()                                                      {
    i2c_ETPP_send_cmd(0x0D);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
    yellowOn();
  }
  void lcdNoBlink()                                                      {
    i2c_ETPP_send_cmd(0x0C);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
  }
  void lcdCursor()                                                      {
    i2c_ETPP_send_cmd(0x0E);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
    yellowOn();
  }
  void lcdNoCursor()                                                      {
    i2c_ETPP_send_cmd(0x0C);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
  }
  void lcdCreateChar(int8_t idx, uint8_t* array)                       {i2c_ETPP_create_char(idx, array);}
  void lcdSetCursor(uint8_t pos)                                       {i2c_ETPP_set_cursor(pos % 16, pos / 16);}
  void lcdWrite(int8_t digit)                                          {i2c_ETPP_send_char(digit);}
  void lcdPrint(char const      msg[])                                 {i2c_ETPP_send_string(msg);}

// ----------

#elif defined(SERIAL_LCD)

  void lcdCmd(byte arg)                                                {lcd.write(0xFE); lcd.write(arg); lcdDelay();}
  void lcdClear()                                                      {lcdCmd(0x01);}
  void lcdBlink()                                                      {lcdCmd(0x0D); yellowOn();}
  void lcdNoBlink()                                                    {lcdCmd(0x0C);}
  void lcdCursor()                                                     {lcdCmd(0x0E); yellowOn();}
  void lcdNoCursor()                                                   {lcdCmd(0x0C);}
  void lcdCreateChar(int8_t idx, uint8_t* array)                       {}
  void lcdSetCursor(uint8_t pos)                                       {if (pos>15) pos += 48; lcdCmd(128+pos);}
  void lcdWrite(int8_t digit)                                          {lcd.write(digit);}
  void lcdPrint(char const msg[])                                      {lcd.print(msg);}

// ----------

#else                                                                  // next set of functions work for LiquidCrystal and SRLCD libraries

  void lcdClear()                                                      {lcd.clear();}
  void lcdBlink()                                                      {lcd.blink();  yellowOn();}
  void lcdNoBlink()                                                    {lcd.noBlink();}
  void lcdCursor()                                                     {lcd.cursor(); yellowOn();}
  void lcdNoCursor()                                                   {lcd.noCursor();}
  void lcdCreateChar(int8_t idx, uint8_t* array)                       {lcd.createChar(idx, array);}
  void lcdSetCursor(uint8_t pos)                                       {lcd.setCursor(pos % 16, pos / 16);}
  void lcdWrite(int8_t digit)                                          {lcd.write(digit);}
  void lcdPrint(char const msg[])                                      {lcd.print(msg);}

#endif // LCD_ETPP

// --------------------------------

void lcdPrint     (uint8_t pos, char const      msg[])                 {lcdSetCursor(pos); lcdPrint(msg);}
void lcdPrintProg (uint8_t pos, prog_char const msg[])                 {lcdSetCursor(pos); lcdWriteProg(msg);}

void lcdPrintProg0(prog_char const line[])                             {lcdPrintProg( 0, line);}
void lcdPrintProg1(prog_char const line[])                             {lcdPrintProg(16, line);}
void lcdPrintProg2(prog_char const line0[], prog_char const line1[])   {lcdPrintProg0(line0); lcdPrintProg1(line1);}

void lcdWrite(uint8_t pos, int8_t digit)                               {lcdSetCursor(pos); lcdWrite(digit);}

void lcdWriteProg(prog_char const msg[]) {
  char* lptr = lineBuffer;
  char  last = -1;
  for (int8_t i = 0, k = 0; last != '\0' && k < 16; i++) {
    last = pgm_read_byte_near(msg + i);
    if (last <= 16) {
      // a char ch < 16 represent empty strings of length ch - this saves a lot of memory
      for (int8_t j = 0; j < last && k < 16; j++) {*lptr++ = ' '; k++;}
    }
    else {
      *lptr++ = last;
      k++;
    }
  }
  *lptr++ = '\0';

  lcdPrint(lineBuffer);
}

// --------------------------------

void format      (uint8_t pos, int val)                          {format      (pos, val, 4,   0);}
void format      (uint8_t pos, int val, int8_t len, int8_t prec) {formatNum   (pos, val, len, prec, false);}

void formatSigned(uint8_t pos, int val)                          {formatSigned(pos, val, 4,   0);}
void formatSigned(uint8_t pos, int val, int8_t len, int8_t prec) {formatNum   (pos, val, len, prec, true);}

void formatNum   (uint8_t pos, int val, int8_t len, int8_t prec, bool sgn) {
  char temp[len+1];
  int8_t low = 0;
  if (sgn) {
    temp[0] = (val > 0)? '+' : (val < 0)? '-' : ' ';
    low++;
  }
  val = abs(val);
  for (int8_t i = len-1; i >= low; i--) {
    if (len-i == prec) {
      temp[i] = '.';
    }
    else {
      temp[i] = '0' + (val % 10);
      val    /= 10;
    }
  }
  if (val > 0) { // overflow
    for (int8_t i = len-1; i >= 0; i--) {
      temp[i] = '#';
    }
  }
  temp[len] = '\0';
  if (pos < 32) {
    lcdPrint(pos, temp);
  }
  else {
    lcdPrint(temp);
  }
}

void formatScaledSym(uint8_t pos, int val) {formatScaled(pos, val, true);}
void formatScaledAbs(uint8_t pos, int val) {formatScaled(pos, val, false);}

void formatScaled   (uint8_t pos, int val, boolean sym) { // sym: -100..+100, !sym=0..100
#ifdef GRAPHIC
  if (graphMode) {
    val = (val - 1000)/10;
    formatGraphic(pos, val);
  }
  else
#endif
  {
    if (sym) {
      val = (val - 1500)/5;
    }
    else {
      val = (val - 1000)/10;
    }
    formatSigned(pos, val);
  }
}

// --------------------------------

#ifdef GRAPHIC
static int8_t  charsInitialized;    // chars for servo signals are initialized
static byte    bars[8] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x15, };

void formatGraphic(uint8_t pos, int8_t val) { // 0..100
  if (charsInitialized != 1) {
    charsInitialized = 1;

    static byte less[8] = { 0x00, 0x04, 0x0C, 0x1C, 0x0C, 0x04, 0x00, 0x15, };
    static byte grt [8] = { 0x00, 0x04, 0x06, 0x07, 0x06, 0x04, 0x00, 0x15, };

    byte pattern = 0x10;
    for (int8_t i = 0; i <= 5; i++) {
      for (int8_t j = 0; j < 7; j++) {
        bars[j] = pattern;
      }
      lcdCreateChar(i, bars);
      pattern >>= 1;
    }
    lcdCreateChar(6, less);
    lcdCreateChar(7, grt);
  }  
  
  static char bar[4];
  for (int8_t i = 0; i < 4; i++)    { bar[i] = 5; }
  
  if      (val < -100 || val > 200) { bar[0] = 6; bar[3] = 7; }    // invalid
  else if (val <    0)              { bar[0] = 6; }                // <...
  else if (val >= 100)              { bar[3] = 7; }                // ...>
  else                              { bar[val/25] = (val%25)/5; }  // ..|.

  lcdSetCursor(pos);
  for (int8_t i = 0; i < 4; i++) {
    lcdWrite(bar[i]); 
  }
}

#if defined(TEXTSTAR_GRAPHIC)
void formatBar(uint8_t len, int8_t val) { // 0..100
  if (charsInitialized != 2) {
    charsInitialized = 2;

    bars[0] = 0x15;
    for (int8_t j = 1; j < 7; j++) {
      bars[j] = 0;
    }
    lcdCreateChar(0, bars);

    bars[3] = 0x10;
    bars[4] = 0x10;
    lcdCreateChar(1, bars);
    
    byte pattern = 0x10;
    for (int8_t i = 2; i <= 6; i++) {
      for (int8_t j = 2; j < 6; j++) {
        bars[j] = pattern;
      }
      lcdCreateChar(i, bars);
      pattern >>= 1;
      pattern |= 0x10;
    }
  }

  int act = (val * len * 6) / 100;
  for (int8_t i = 0, low = 0; i < len; i++, low += 6) {
    char ch = 0;
    if (act >= low + 6) {
      ch = 6;
    }
    else if (act >= low) {
      ch = act - low;
    }
    lcdWrite(ch);
  }
}
#endif // TESXTSTAR

// ----------

static uint8_t angleChars[4][8];

int8_t round8(float val) {return (int8_t) (val + 0.5f);}

void drawLine(int8_t x0, int8_t y0, int8_t x1, int8_t y1) {
  // Bresenham again...
  
  int8_t dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
  int8_t dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
  int8_t err = dx+dy, e2; /* error value e_xy */
 
  for(;;){  /* loop */
    setPixel(x0, y0);
    if (x0==x1 && y0==y1) break;
    e2 = 2*err;
    if (e2 >= dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
    if (e2 <= dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
  }  
}

void setPixel(int8_t x, int8_t y) { 
  // x: -10..+9, y: -3..+4
  
  if (x >= -10 && x < 10 && y >= -3 && y < 5) {
    int8_t nrow = 4 - y;   // 0..7
    int8_t ncol = x + 10;  // 0..19
    int8_t idx  = ncol / 5;
    int8_t pcol = ncol % 5;
    
    angleChars[idx][nrow] |= (0x10 >> pcol);
  }
}
#endif // GRAPHIC

// --------------------------------

void drawAngles() {
  for (int8_t index = 0; index < 2; index++) {
#ifdef GRAPHIC
    if (graphMode) {
      // clear char arrays
      uint8_t *cptr = angleChars[0];
      for (int8_t i = 0; i < 32; i++) {
        *cptr++ = 0;
      }
    
      int8_t  charOffset = index * 4;
      int16_t alpha      = - angle[index] MM_VAL;
      if (alpha >= -360 && alpha <= 360) {
        float ang  = radians(alpha);
        float sinA = sin(ang);
        float cosA = cos(ang);
        
      //  if (abs(sinA) > 0.25f) {
      //    angleChars[0][4] = 0x14;  // mark horizon
      //    angleChars[3][4] = 0x05;
      //  }
        
        int8_t x0 = round8(- cosA * 10.0f);
        int8_t y0 = round8(- sinA * 10.0f);
        drawLine(x0, y0, -x0, -y0);
        
        int8_t x1 = round8(- sinA * 3.0f);
        int8_t y1 = round8(  cosA * 3.0f);
        drawLine(0, 0, x1, y1);
      }
      else {
        angleChars[1][4] = 0x06;  
        angleChars[2][4] = 0x18;  
      }

      for (uint8_t i = 0; i < 4; i++) {
        lcdCreateChar(charOffset + i, angleChars[i]);
      }      
      lcdSetCursor(16 + 2 + index * 8);
      for (uint8_t i = 0; i < 4; i++) {
        lcdWrite(charOffset + i);
      }
    }
    else 
#endif // GRAPHIC
    {
      formatSigned(16 + 2 + index * 8, angle[index] MM_VAL);
    }
  }
#ifdef GRAPHIC
  charsInitialized = 0;  
#endif
}

// --------------------------------

void printEsc(uint8_t pos, uint8_t code) {
  if (code >= 20) {
    if (code == 30) {
      code = (vers < 20)? 20 : 25;  // 30 -> 20, 25 (TRI)
    }
    else if (code == 31 || code == 32) {
      code -= (vers < 20)? 11 : 7; // 31 -> 20, 24, 32 -> 21, 25 (BI)
    }
    
    formatScaledSym(pos, servo[code-20] MM_VAL);  // servo signals - symmetrical
  }
  else {
    formatScaledAbs(pos, motor[code] MM_VAL);
  }
}

// --------------------------------

void printAuxState() {
  uint16_t mode = actMode[auxState];

  lcdSetCursor(16);
  char stat = (mode & 0x8000)? '@' : ' ';
  lcdWrite(stat);
  
  for (uint8_t i = 0; i < 12; i++) {
    if (i > 0 && (i % 3) == 0) {
      lcdWrite(' ');
    }

    char ch = (mode & auxMasks[i])? '*' : DOT;
    lcdWrite(ch);
  }
}

// --------------------------------

#ifdef EMULATE_TEXTSTAR
void emulateLcd() {
  static char lcd_char[] = { 'C', 'D', 'B', 'A' };
  
  lcdPrintProg1(emuLcdMsg);
  delay(2000);
  lcdClear();
  
  uint8_t bsize;
  uint8_t state = 0;
  while (true) {
    if (serialAvailable()){
      uint8_t next = serialReadByte();
      if (state == 0) {
        if      (next == 0xfe) {state = 1;       }  // control character escape sequence
        else if (next == 0x0c) {lcdClear();      }
        else if (next == 0x0d) {lcdSetCursor(16);}
        else                   {lcdWrite(next);  }
      }
      else if (state == 1) {
        if      (next == '+') {state = 2;} // cursor BlinkMode ON
        else if (next == 'L') {state = 3;} // set cursor position
        else if (next == 'b') {state = 4;} // bar graphic
        else                  {state = 0;} // ignore
      }
      else if (state == 2) {
        lcdBlink();
        state = 0;
      }
      else if (state == 3) {
//        lcdSetCursor((next == 0x01)? 0 : 16);
        if (next == 0x01) {lcdSetCursor( 0);}
        else              {lcdSetCursor(16);}
        state = 0;
      }
      else if (state == 4) {
        // next is size of output field
        bsize = next;
        state = 5;
      }
      else if (state == 5) {
        // next is supposed to be a value in percent (?)
#if defined(GRAPHIC) && defined(TEXTSTAR_GRAPHIC)
        formatBar(bsize, next);
#else       
        formatNum(100, (int8_t) next, bsize, 0, true);
#endif
        state = 0;
      }
    }
    else {
      // in order to allow switching of Telemetry displays, the refresh rate of
      // the telemetry should be set to 1 Hz in the MWC sketch (otherwise, it
      // ignores sent characters because the Serial is always busy).

      flashYellow();
      int8_t key = getKeyPressed();
      if (key == KEY_SELECT) {
        break;
      }
      else if (key >= 0) {
//        char                          ch = 'B';   // LCD_VALUE_DOWN
//        if      (key == KEY_LEFT)   { ch = 'A'; } // LCD_VALUE_PREV
//        else if (key == KEY_RIGHT)  { ch = 'C'; } // LCD_VALUE_NEXT
//        else if (key == KEY_UP)     { ch = 'D'; } // LCD_VALUE_UP
//        else if (key == KEY_SELECT) { break; }

        char ch = lcd_char[key];
        serialWrite(ch);                          // key down
        waitForKeyUp();
        serialWrite(ch+32);                       // key up
        flashRed();
      }
      delay50();
    }
  }
  waitForKeyUp();
}
#endif

