/*
 * ----- (lcd_config.h) Configuration: -----
 */

// Four different Options for connecting a LCD:
// (1) Standard 16x2 HD44780 as connected in the DFRobot Keypad Shield (default, no specific option required)
// (2) HD44780 connected via a shift register using the ShiftRegLCD library (uncomment #define SRLCD)
// (3) Eagle Tree PowerPanel connected via I2C (uncomment #define LCD_ETPP)
// (4) Sparkfun Serial LCD connected via a Soft Serial connection (uncomment #define SERIAL_LCD)
// Make sure you uncomment at most ONE option.

// use ShiftRegLCD lib instead of LiquidCrystal lib
// [footprint: decreases ~200 bytes]
//#define SRLCD     

// use Eagle Tree Power Panel (no libraries, via I2C).  Hookup White to Ground, Red to +5V, Yellow to A4, Brown to A5
// (code contributed by Danal Estes)
// [footprint: decreases ~900 bytes]
//#define LCD_ETPP

//  Eagle Tree Power Panel  http://www.eagletreesystems.com/MicroPower/9.htm
//  The Power Panel is a physically small & light weight, 2x16, transflective (readable in sunlight) LCD that interfaces via i2c bus (two wire).
//  White wire  to Ground
//  Red wire    to +5V
//  Yellow wire to A4
//  Brown wire  to A5
//  ETPP code contributed by Danal Estes

// use Sparkfun Serial LCD (via SoftwareSerial lib), Data pin of the display has to be connected to pin SERIAL_LCD_TX
// [footprint: decreases ~100 bytes]
//#define SERIAL_LCD

// ----------

// include routines to show some output as simple character graphic (for servo signals and angles) 
// [footprint: ~2.5 kbytes]
#define GRAPHIC

// ----------

// if defined, allows dimming of LCD backlight intensity (Pin 10)
// supported by KeypadShield only (?)
// [footprint: ~250 bytes]
#define BACKLIGHT

#define BACKLIGHT_PIN 10

// ----------

// Emulation of a TextStar Serial LCD to allow use of the MWC built-in configuration. You have to enable both the
// LCD_CONF and LCD_TEXTSTAR constants in the MWC config.h file.
// [footprint: ~520 bytes]
#define EMULATE_TEXTSTAR

// Displays Bar Charts in Telemetry Displays (requires GRAPHICS Option)
// [footprint: ~230 bytes]
#define TEXTSTAR_GRAPHIC

// ----------

// Three different option for input devices:
// (1) Cascading buttons as connect in the DFRobot Keypad shield (default, no specific option required)
// (2) Separate buttons connected via a resistor against VCC when open and against ground when closed (option HWKEYS)
// (3) Two way analog Joystick, plus an optional SELECT key (option JOYSTICK)
// Cases (1) and (2) are mutual exclusive, (3) can be used parallel to (1) xor (2). Key inputs have priority over Joystick inputs.
// The four buttons (LEFT, RIGHT, UP, DOWN) are mandatory, the fifth button (SELECT) is more or less optional.
// The actual state of the input devices can be displayed on the screen when the CHECK_INPUT option is active. Press LEFT for more
// than one second in the start state to enter the CHECK_INPUT state, press RESET to leave this state.

// use Hardwarekeys connected to Ports
// [footprint: ~160 bytes]
//#define HWKEYS

// Input pins of the hardbuttons: (used when HWKEYS is defined)
#define HW_RIGHT     10
#define HW_UP         9
#define HW_DOWN      11
#define HW_LEFT       8
#define HW_SELECT     6

// Analog Input pin for cascading buttons (comment this line if you are not using cascading buttons)
#define KEY_PIN      A0

// ----------

// Use an analog Joystick (e.g., Sparkfun Thumb Joystick) instead of Keys (experimental!)
// Joystick and Buttons can be used simultaneously. The center position of the Joystick
// is calibrated during startup, so leave the stick in central position when switching on.
// [footprint: ~280 bytes]
#define JOYSTICK

#define JS_HOR_PIN   A1
#define JS_VERT_PIN  A2
// uncomment the following line to enable the joystick select key 
// The select key has to be connected to GND when pressed! 
//#define JS_SELECT_PIN A3

// Configuration of the Axes
#define JS_HOR_SIGN  +1 
#define JS_VERT_SIGN -1
#define JS_DEADBAND  48

// ----------

// includes a hidden state to check the key and joystick input. 
// activated by pressing the RIGHT key in state 0 for more than a second.
// [footprint: ~170 bytes]
#define CHECK_INPUT

// ----------

// include routines to show the state of the application with three leds 
// (red: error, yellow: busy/waiting for input, green: logging)
// [footprint: ~300 bytes]
#define STATUS_LEDS

// Caution: Default Pins of LEDs and Joystick input overlap
#define RED_LED      A5
#define YELLOW_LED   A4
#define GREEN_LED    A3

// Battery Level Warning Levels:
//
// ----------RED------->|             |<------GREEN---------
//                      |             |
//  VBATLEVEL3 ... VBATLEVEL2 ... VBATLEVEL1 ... VBATLEVEL0
//      |                                            |
//      |<------------------YELLOW------------------>|
//
// the battery status is presented only in the corresponding state ("3 - Rates" -> "Batt Power Alarm")
// (requires the STATUS_LED Option)

#define VBATLEVEL0 114  // 11.4V - 3.8V per cell @ 3S [above: green,        below: green/yellow]
#define VBATLEVEL1 111  // 11.1V - 3.7V               [above: green/yellow, below: yellow      ]
#define VBATLEVEL2 108  // 10.8V - 3.6V               [above: yellow,       below: yellow/red  ]
#define VBATLEVEL3 105  // 10.5V - 3.5V               [above: yellow/red,   below: red         ]

// ----------

// track minimum and maximum values of certain values
// [footprint: ~260 bytes]
#define MINMAX

#ifdef MINMAX
  #define MM_ACT [0]
  #define MM_DIM [3]
  #define MM_VAL [minMaxMode]
#else
  #define MM_ACT
  #define MM_DIM
  #define MM_VAL
#endif

// ----------

// :)
#define CREDITS

// simulates a serial connection, SELECT key in state "ESC Out" changes the configuration
//#define TEST

// if defined, dumps the contents of the read buffer hexadecimal on the LCD when an invalid frame is read
// (useful for debugging the serial interface)
// [footprint: ~310 bytes]
#define DUMP_BUFFER

// ----------------

// Number of hardbuttons:
#define NUM_KEYS     5

// Identifiers for the five hardbuttons:
#define KEY_RIGHT    0
#define KEY_UP       1
#define KEY_DOWN     2
#define KEY_LEFT     3
#define KEY_SELECT   4
#define STATUS       5         // Last column in the transition table is reserved for status flags

#define KEY_DECR     KEY_UP    // Specifies which key is used to decrement values (KEY_UP or KEY_DOWN)

// RC Channel Aliases:
#define ROLL         0
#define PITCH        1
#define YAW          2
#define THROTTLE     3
#define AUX1         4
#define AUX2         5
#define AUX3         6
#define AUX4         7

// ----------

// Baudrate Settings 
#define DEFAULT_BAUDRATE 8

// when defined, the DEFAULT_BAUDRATE is used permanently, otherwise the baudrate is stored to and 
// restored from EEPROM. In both cases, the user can change the baud rate manually in the 'Settings' state
//#define USE_DEFAULT_BAUDRATE 

// ----------

// Number of User Profiles [default: 12, depends on EEPRom size (1024 on ATMega328, 512 on ATMega168, maximum = (EEPROM_size - 8)/84)]
#define NUM_PROFILES 12   // default - maximum on an ATMega328 (12 * 84 + 8 = 1016)
//#define NUM_PROFILES 6    // maximum on an ATMega168 (6 * 84 + 8 = 512)

