/*
 * LCD_Config v4.4 (for Arduino 1.0) BETA
 * ---------------
 *
 * last change: 11.03.2012
 *
 * Allows configuration of the MultiWii software via a serial connection.
 * Requires a LCD display and at least four buttons (left, right, up, down)
 * or an analog joystick. The SELECT button is for both configurations optional.
 *
 * Options:
 *   Status LEDs (three LEDs to present status information)
 *   Setting of LCD backlight intensity (if supported)
 *   Use of ShiftRegLCD lib instead of LiquidChrystal lib, or
 *   Use of EagleTree PowerPanel, or
 *   Use of Sparkfun Serial LCD
 *   Emulation of a Textstar LCD to configure the copter via the MWC-internal LCD_CONF functionality
 *   Use hardware buttons instead of buttons connected via cascading resistors
 *   Use of an analog joystick instead of buttons
 *   Graphic output (simple character graphic for angles and servo signals)
 *   MinMax mode (store und display minimum/maximum values of certain fields)
 *   [Test] (for standalone testing without a MWC board connected)
 *   [CHECK_INPUT] (presents the state of the input pins - press LEFT in start state for >1 sec, exit by RESET)
 *   [DUMP_BUFFER] (dumps the bytes in the serial buffer to the LCD (in hexadecimal) when an invalid
 *   frame is received (useful for debugging))
 *
 * IMPORTANT: Due to a bug in the way Arduino tread #include commands, all #includes are now in
 * a separate block at the beginning of this Sketch. When configuring the Sketch to your hardware, 
 * you have to select your set of options and the libraries these options require separately.
 * Other way round, you have to comment/uncomment lines in two locations. Annoying, yes, but it saves
 * memory since the unused libraries are linked to the executable otherwise.
 *
 * ----------
 *
 * To enter the Textstar Mode, press RIGHT immediately after startup.
 * The MWC board enters the configuration loop when the Throttle stick is at MIN and the YAW and PITCH
 * sticks are at MAX. Be careful, these stick positions are very close to arming. Always move the
 * PITCH stick first, then YAW.
 *
 * ----------
 *
 * Graphic Mode is toggled either by pressing the SELECT key or by pressing the RIGHT key longer
 * than 1 sec. Min/Max Mode is toggled by pressing the RIGHT key shorter than 1 sec.
 *
 * ----------
 *
 * Authors: yacco, 
 *          calli        (HWKEYS, ShiftRegLCD), 
 *          Danal Estes  (ETPP [Eagle Tree Power Panel] support)
 *          dynai        (early version of EMULATE_TEXTSTAR)
 *
 * changes to 1.25: 
 *   display values of gyro and acc changed to signed
 *   show minimum/maximum values of sensor data and servo signals
 * changes to 1.28:
 *   wait state after startup, no serial communication during initialization of the MWC
 *   removed min/max values
 *   added LED status signals
 * changes to 1.30:
 *   bugfix in playback routine, added timeout in playback
 * changes to 1.31:
 *   minor bugfixes, added setting of backlight
 * changes to 1.32:
 *   removed KEY_SELECT column in next_state, backlight and waitForInit optional
 * changes to 1.33:
 *   check trailing 'A' char in serial communication to detect frame errors, 
 *   optimization of format routines (merged two routines, invalid values are displayed as ###)
 *   bugfix in logging (logMode=-1 was stored in flash, so logging was permanently disabled)
 *   KEY_LEFT cancels reading when no serial connection is present (important to reset wrong settings!) 
 * changes to 1.5:
 *   reimplemented Min/Max-Mode (found it useful again), 
 *   fixed compiling problems for certain combinations of options 
 * changes to 1.6:
 *   adaptation to MWC version 1.preter6 (new config "Flying Wing", new data format)
 *   changed status display (state 0 - Config)
 *   bug 'logMode=-1' finally solved
 *   Version has been set to 2.0 to disambiguate between MWC version and LCD_Conf version.
 * changes to 2.0 beta: 
 *   swapped motors 4 and 5 in 'Esc Out' for versions 1.5 and 1.6prebis
 *   changed behaviour of UP and DOWN for consistency (UP decrements, DOWN increments values)
 *   added version V1.6 
 *   output logfiles in LogView format [OPTIONAL] (file numbers starting from 500, extension ".lvw")
 *   added JOYSTICK option to replace input buttons by an analog joystick [experimental]
 *   display 'invalid' values in graphics
 *   heavy optimizations (!) due to space restrictions
 *   removed WAIT_FOR_INIT option, changed start state to 31 instead 
 * changes to 2.0:
 *   bugfix in deserialize() with v1.6
 * changes to 2.01:
 *   Joystick handling was improved (calibration, use of pullups, can be used in parallel to buttons)
 *   Added 'Play LogFile' state
 *   Easter egg function to display internal configuration :)
 *   BUGFIX: RC Rate multiplied by factor 2 for MWC Versions >v1.5, minimum of 20% for RC Rate (0% makes no sense)
 *   Added random offsets in TEST mode
 * changes to 2.1:
 *   Cancel in Logfile Playback (press LEFT_KEY)
 *   Settings of "Baud Rate", serialRead() has been adapted to work with different COM speeds
 *   Greater flexibility in changing the Settings
 *   (Log) Delay Time was moved to Settings state and is used as the time interval of polling the MWC board
 * changes to 2.2:
 *   Changed internal representation of state machine and strings
 *   Support for v1.7pre (I-/P-Level instead of AutoLevelStrength, trigger calibration, 8 instead of 5 receiver
 *   channels, configuration of activation modes for Aux1/Aux2, ...)
 *   Software Version is now stored in Logfile
 *   APC220 Option (disables APC220 module via pin D2 during playback of logfiles, experimental)
 * changes to 2.3:
 *   Swapped Angle and Baro/Comp/Cycle states
 *   Added 'Config' parameter to LogFiles, show Config and Version in selectLogFile
 *   Use of blinking Cursor instead of Arrows in Enter mode
 *   LogView mode: corrected order of receiver channels
 * changes to 2.4:
 *   "Roll/Pitch" state - simultaneous setting of Roll and Pitch PID values
 * changes to 2.5:
 *   Changed representation of configIds
 *   'pos' parameter in lcd output instead of (col, row), saved ~250 bytes of flash
 *   Support for v1.7prebis (replaced v1.7pre)
 *   CHECK_INPUT option to visualize the joystick input (hidden state, press LEFT in start for >1 sec, exit by RESET)
 *   BATTERY_LEVEL option (experimental), visualisation on LCD and via Status LEDS, 4 warn levels can be defined
 *   Optimized presentation of ESC/Servo signals (saved ~750 bytes)
 * changes to 2.6:
 *   Added consistency check in serialRead, removed strange behaviour after framelength errors.
 *   Improved error handling when communication fails (more detailed description than "Inv. FrameLength")
 *   Show MWC version in start state
 *   Show current values in "Settings" state, added blinking cursor when entering values
 *   Number of Profiles can be configured (constant NUM_PROFILES)
 * changes to 2.7:
 *   Removed BATTERY_LEVEL problem in conjunction with APC220
 *   Flashing in battery status display to indicate warn levels
 *   Bugfix in setting of delay time
 * changes to 2.8:
 *   KEY_PIN can be undefined if no cascading Buttons are used
 *   Changed default values for Version (highest version) and Baudrate (115200)
 *   Added Option "Dump Buffer" to show the contents of the data buffer when an invalid frame is received (for debugging)
 *   DEFAULT_BAUDRATE can be configured, USE_DEFAULT_BAUDRATE ignores stored Baudrate params (Baudrate can still be adjusted manually)
 *   Bugfix in state "Enter R/C Rate / Expo" (no action on RIGHT)
 *   Bugfix concerning Battery level (removed leading extra byte)
 * changes to 2.9:
 *   Removed JS_INV_SELECT_KEY (Joystick Select key has to be connected to GND when closed).
 *   Bugfix in state "Enter Level PID" (no action on RIGHT)
 * changes to 3.0:
 *   Eagle Tree Power Panel (Option LCD_ETPP) support (setting of backlight not supported)
 *   Sparkfun Serial LCD (Option SERIAL_LCD) support (graphic mode not supported)
 * changes to 3.1:
 *   changed Serial LCD from SoftwareSerial lib to NewSoftSerial lib (see http://arduiniana.org/libraries/NewSoftSerial/)
 *   added EMULATE_SERIAL_LCD option to use the LCD_Config as a Serial LCD (option LCD_CONF has to be enabled in MWC software).
 *   - To enter the emulation state press KEY_RIGHT during the initial 'Reading' state. 
 *   - This feature allows you to use the LCD_Config with MWC versions not supported by LCD_Config (i.e., the developer version).
 *   bugfix in restoreParams()
 *   in "Reading..." state, the characters '<' and '>' indicate which keys are currently selectable. ('<' to cancel, '>' to enter
 *   the fabulous 'Serial LCD Emulation' mode.)
 *   extracted option section to separate Header file, the user should ONLY edit this file and the comments in the #include lines below
 * changes to 3.2
 *   replaced EMULATE_SERIAL_LCD by EMULATE_TEXTSTAR
 *   added support for MWC Version 1.8
 *   therefor, removed support for MWC Version < 1.7 (!)
 *   removed BATTERY_LEVEL option (now mandatory), added support for PowerMeter
 * changes to 4.0pre:
 *   formatBar() for Textstar Emulation Mode
 *   Bugfix in resetProfile()
 *   Bugfix in playbackLog()
 *   Optimized write access to EEPROM
 * changes to 4.0:
 *   Extended lenght of Altitutde field to 6 characters
 * changes ot 4.0p1:
 *   Support for v1.9 (+GPS)
 *   Rearranged Substates of State '0 - Config'
 *   Bugfix 'Enter PowerTrigger value'
 *   Bugfixes in Profile Management
 *   Show 'Debug' Values (Substate of '0 - Config')
 * changes to 4.1:
 *   Support for Arduino 1.0
 * changes to 4.2:
 *   Bugfixes, Changes in handling of input keys and serial communication
 * changes to 4.3:
 *   Removed Options "LOGGING", "APC220" and "STANDALONE"
 *   Splitted source file
 *   Bugfix in DrawAngle() routine (only in Graphics mode)
 *   New layout for Profiles, reduced max. number of profiles (profile size increased from 48 bytes to 84 bytes)
 *   Extended character set for Profile names
 *   Support for MWC 2.0, i.e.:
 *   -renamed receiver channels 6 and 7 to Aux3 and Aux4
 *   -6 new configurations: OctFlat+, OctFlatX, Airplane, Heli_120, Heli_90, VTail4 (Airplane and Heli do not seem to be implemented in MWC yet)
 *   -number of servos increased to 8, adapted servo indexes for TRI and BI
 *   -added status variable "i2c_errorCount" (new substate of "Sensors" state)
 *   -new activation modes: PassThru, Headfree, Beeper
 *   -new scheme to enter activation modes
 *
 * overall footprint (on an Arduino Uno, using the LiquidCrystal lib and cascading buttons): 
 *  ~17862 bytes with all options enabled using a Keypad Shield for IO
 *  ~13770 bytes with all options disabled (GRAPHIC, BACKLIGHT, JOYSTICK, CHECK_INPUT, EMULATE_TEXTSTAR, STATUS_LEDS, MINMAX)
 *
 * (the maximum image size uploadable via bootloader on a AtMega328 is ~28670 bytes)
 */

/*
 * ----- #include Section -----
 */
 
// include Constants/Macros for Options 
#include "LCD_Config.h"

// for persistence of SW Version, DelayTime and Profiles (mandatory)
#include <EEPROM.h>

// defines the PROGMEM macro to move constant variables from SRAM to flash memory (mandatory)
#include <avr/pgmspace.h> 

// comment this line when NOT using a regular HD44780
#include <LiquidCrystal.h>

// uncomment this line when using ShiftReg LCD (Option SRLCD)
#ifdef SRLCD
//#include <ShiftRegLCD.h>      
#endif

// uncomment this line when using a Serial LCD (Option SERIAL_LCD)
#ifdef SERIAL_LCD
//#include <SoftwareSerial.h>
#endif

// --------------------------------

// Workaround for "insertion of forward declarations in excluded blocks" problem
// Don't move this line around.
void setup();     

// --------------------------------

/*
 * ----- Configurations -----
 */
 
// (1) LCD:
// --------

// 4 different Hardware Options exist: 
// Regular HD44780 via LiquidCrystal lib, 
// via Shiftreg Library, 
// Sparkfun Serial LCD via SoftSerial, 
// Eagle Tree Power Panel via I2C

// HD44780 LCD via the ShiftReg Library
#if defined(SRLCD)
  ShiftRegLCD                lcd(3, 2, 4, 2);    // (Datapin, Clockpin, Enablepin or TWO_WIRE [, Lines [, Font]]])
  #define INIT_LCD                               // no begin in ShiftRegLCD...

// Eagle Tree Power Panel
#elif defined(LCD_ETPP)
  #define INIT_LCD           i2c_init(); i2c_ETPP_init();
  #undef BACKLIGHT                               // backlight not supported

// Sparkfun Serial LCD
#elif defined(SERIAL_LCD)
  #define SERIAL_LCD_TX      2                   // emulates serial port with TX on D2
  #define LCD_DELAY          4                   // conservative, 2 actually works
  #define INIT_LCD           pinMode(SERIAL_LCD_TX, OUTPUT); lcd.begin(9600);

  #undef GRAPHIC                                 // custom characters not supported

  SoftwareSerial             lcd(0, SERIAL_LCD_TX);
  void lcdDelay()            {delay(LCD_DELAY);}  // delay for transmission

// Regular HD44780 via Liquid Crystal Library (e.g., DFRobot Keypad Shield) - default
#else
  LiquidCrystal              lcd(8, 9, 4, 5, 6, 7); // rs, enable, d4, d5, d6, d7, 
  #define INIT_LCD           lcd.begin(16, 2); 
#endif

#if defined(LCD_ETPP)
  #define LEFT_ARROW        "\074"
  #define RIGHT_ARROW       "\076"
  #define DOT               '.'
#else
  #define LEFT_ARROW        "\177"
  #define RIGHT_ARROW       "\176"
  #define DOT               165
#endif  

// (2) LCD Backlight: (if supported by hardware)
// ------------------

#ifdef BACKLIGHT
  static uint8_t            backlight;   // backlight intensity (0..255)

  #define BL_MSG            "Backlight"
  #define NUM_SETTINGS      3
  #define INIT_BACKLIGHT    ee_ptr = 5; backlight = 255-readEEPROM(); setBacklight(backlight);

  void setBacklight(uint8_t arg) {
  #ifdef SERIAL_LCD
    uint8_t bl = 157 - max(arg/5, 29); 
    lcdCmd(bl);
  #else
    analogWrite(BACKLIGHT_PIN, arg);
  #endif
  }

#else
  #define BL_MSG
  #define NUM_SETTINGS      2
  #define INIT_BACKLIGHT
#endif

// (3) Joystick:
// -------------

#ifdef JOYSTICK
  static int16_t jsHorCtr, jsVertCtr;
  static int16_t jsHorAct, jsVertAct;

  #define INIT_JOYSTICK     pinIn(JS_HOR_PIN); pinIn(JS_VERT_PIN); \
                            for (int8_t i = 0; i < 8; i++) {jsHorCtr += analogRead(JS_HOR_PIN); jsVertCtr += analogRead(JS_VERT_PIN); delay(25);} \
                            jsHorCtr /= 8; jsVertCtr /= 8;
#else
  #define INIT_JOYSTICK
#endif

#ifdef JS_SELECT_PIN
  #define INIT_JS_SELECT    pinIn(JS_SELECT_PIN);
#else
  #define INIT_JS_SELECT
#endif

// (5) Status LED:
// ---------------

#ifdef STATUS_LEDS
  #define INIT_LEDS         pinOut(RED_LED); pinOut(YELLOW_LED); pinOut(GREEN_LED);

  void red   (boolean a)    {digitalWrite(RED_LED,    a);}
  void yellow(boolean a)    {digitalWrite(YELLOW_LED, a);}
  void green (boolean a)    {digitalWrite(GREEN_LED,  a);}

  void yellowOn ()          {digitalWrite(YELLOW_LED, true); }
  void yellowOff()          {digitalWrite(YELLOW_LED, false);}

  void blinkLed(uint8_t led, uint16_t del) {for (int8_t i=1; i >= 0; i--) {digitalWrite(led, i); delay(del);}}
#else
  #define INIT_LEDS

  #define red(a)
  #define yellow(a)
  #define green(a)

  #define yellowOn()
  #define yellowOff()

  void blinkLed(uint8_t led, uint16_t del) {delay(del+del);}
#endif // STATUS_LEDS

void flashRed()             {blinkLed(RED_LED,    10);}
void flashYellow()          {blinkLed(YELLOW_LED,  5);}
void flashGreen()           {blinkLed(GREEN_LED,   5);}
void blinkConfirm()         {blinkLed(GREEN_LED, 250);}
void blinkWarning()         {for (uint8_t i = 6; i > 0; i--) {blinkLed(RED_LED, 125);}}

// (6) Test Mode:
// --------------

#ifdef TEST
  #define INIT_TEST         lcdPrint(12 + 16, "Test"); randomSeed(4711); config = 1;
#else
  #define INIT_TEST
#endif

/*
 * ----- String Literals and Constants: ----- 
 */

#define VERSION      "4.4"

// Number of States in State Machine
#define NUM_STATES   48

// Configuration State Offset
#define CONF_OFFSET  23

// Number of Configurations: (Unknown, Tri, Quad+, QuadX, Bi, Gimbal, Y6, Hex, Flying Wing, Y4, HexX, Oct, OctFlat+, OctFlatX, Airplane, Heli_120, Heli_90, VTail4)
#define NUM_CONFIGS  18

// Activation Modes: level, baro, compas, camStab, camTrig, arming, gpsHome, gpsHold, passThru, headfree, beeper
#define AUX_STATES    11

// Max. Number of Bytes read from Serial
#define BUFFER_SIZE  192

// cryptic representation of selected compiler options.
// can be visualized at run-time by pressing LEFT in start state for >1 sec.
// useful when testing different sets of options
PROGMEM prog_char metaData[] = VERSION ":" 
#ifdef TEST
  "T"
#endif
#ifdef SRLCD
  "S"
#endif
#ifdef LCD_ETPP
  "P"
#endif
#ifdef SERIAL_LCD
  "C"
#endif
#ifdef BACKLIGHT
  "B"
#endif
#ifdef EMULATE_TEXTSTAR
  "X"
#endif
#ifdef KEYPIN
  "K"
#elsifdef HWKEYS
  "H"
#endif
#ifdef JOYSTICK
  "J"
#endif
#ifdef GRAPHIC
  "G"
#endif
#ifdef MINMAX
  "M"
#endif
#ifdef STATUS_LEDS
  "L"
#endif
  ":"
;

// ----------------

PROGMEM prog_char intro         [] = " LCD Config " VERSION;
PROGMEM prog_char emptyLine     [] = "\020";

#ifdef EMULATE_TEXTSTAR
  PROGMEM prog_char reading     [] = "Reading:\006" LEFT_ARROW RIGHT_ARROW;  // '<>' indicates KEY_LEFT and KEY_RIGHT are active
  PROGMEM prog_char emuLcdMsg   [] = "TEXTSTAR Mode\004";
#else
  PROGMEM prog_char reading     [] = "Reading:\007" LEFT_ARROW;              // '<' indicates KEY_LEFT is active
#endif

PROGMEM prog_char rollPitchYaw  [] = ": Rol Pit Yaw";    
PROGMEM prog_char writeMsg      [] = " Wrote\004Bytes ";
PROGMEM prog_char cancelRead    [] = "Reading canceled";
PROGMEM prog_char wrongMsg      [] = "Wrong\011";
PROGMEM prog_char timeoutError  [] = " Timeout Error\002";
PROGMEM prog_char yesNoMsg      [] = " " LEFT_ARROW ": no, " RIGHT_ARROW ": yes\002";
PROGMEM prog_char delaySuff     [] = "\0100 ms";
PROGMEM prog_char offlineMsg    [] = "Offline";
PROGMEM prog_char haltMsg       [] = " Program halted ";
PROGMEM prog_char profileMsg    [] = " Profile";

PROGMEM prog_char frameError    [][ 7] = { "Length", "Prefix", "Suffix", "Config", "Vers.", };
PROGMEM prog_char profStates    [][ 6] = { "Load", "Save", "Clear", "Reset" }; 
PROGMEM prog_char auxStates     [][ 9] = { "Level", "Baro", "Magnet", "CamStab", "CamTrig", "Arming", "GpsHome", "GpsHold", "PassThru", "Headfree", "Beeper",  }; 
PROGMEM prog_char setMsgs       [][11] = { "Delay Time", "Baud Rate", BL_MSG };

#ifdef MINMAX
  PROGMEM prog_char minMaxDsc   [] = "Act\0Min\0Max";
#endif

#ifdef CHECK_INPUT
  PROGMEM prog_char jsStatusMsg [] = "\002Key\002Hor\002Vert";
#endif  

#ifdef CREDITS
  PROGMEM prog_char credits     [] = { 40, 99, 41, 50, 48, 49, 50, 32, 98, 121, 32, 121, 97, 99, 99, 111, 0 };
#endif

// ----------------

PROGMEM prog_char versions[][5] = {
  "1.8 ", 
  "1.9 ",
  "2.0p",
};

// number of bytes per frame send from the MultiWii board, dependant on version
uint8_t frameLength[] = { 
  116,  // 1.8
  125,  // 1.9
  155,  // 2.0p
};

/* ----- */

/*
#if   defined(TRI)            #define MULTITYPE 1
#elif defined(QUADP)          #define MULTITYPE 2
#elif defined(QUADX)          #define MULTITYPE 3
#elif defined(BI)             #define MULTITYPE 4
#elif defined(GIMBAL)         #define MULTITYPE 5
#elif defined(Y6)             #define MULTITYPE 6
#elif defined(HEX6)           #define MULTITYPE 7
#elif defined(FLYING_WING)    #define MULTITYPE 8
#elif defined(Y4)             #define MULTITYPE 9
#elif defined(HEX6X)          #define MULTITYPE 10
#elif defined(OCTOX8)         #define MULTITYPE 11   //the JAVA GUI is the same for all 8 motor configs 
#elif defined(OCTOFLATP)      #define MULTITYPE 12   //12  for MultiWinGui
#elif defined(OCTOFLATX)      #define MULTITYPE 13   //13  for MultiWinGui 
#elif defined(AIRPLANE)       #define MULTITYPE 14    
#elif defined(HELI_120_CCPM)  #define MULTITYPE 15   // Simple model 
#elif defined(HELI_90_DEG)    #define MULTITYPE 16   // Simple model  
#elif defined(VTAIL4)         #define MULTITYPE 17
#endif
*/

// configuration ids
PROGMEM prog_char configIds[] = { 
  "Unknown\0"   //  0
  "Tri\0"       //  1
  "Quad+\0"     //  2
  "QuadX\0"     //  3
  "Bi\0"        //  4
  "Gimbal\0"    //  5
  "Y6\0"        //  6
  "Hex\0"       //  7
  "Fl.Wing\0"   //  8
  "Y4\0"        //  9
  "HexX\0"      // 10
  "OctX\0"      // 11
  "OctF+\0"     // 12
  "OctFx\0"     // 13
  "Plane\0"     // 14
  "Heli120\0"   // 15
  "Heli 90\0"   // 16
  "VTail"       // 17
};

//                           0                  5                  10                  15
uint8_t configOffsets [] = { 0, 8, 12, 18, 24, 27, 34, 37, 41, 49, 52, 57, 62, 68, 74, 80, 88, 96,  };

// Tri:         Left, Right, Rudder, Rear
// Quad+:       Front, Rear, Left, Right
// QuadX:       Front Left, Front Right, Rear Left, Rear Right
// Bi:          Motor 0, Motor 1, Servo 0, Servo 1 
// Gimbal:      Roll, Pitch
// Y6:          Up (Left, Mid/Rear, Right), Down (Left, Mid, Right)
// Hex:         Front (Left, Mid, Right), Rear (Left, Mid, Right)
// Flying Wing: Servo 0-3 
// Y4:          Front Left, Front Right, Rear Up, Rear Down
// HexX:        Left (Front, Mid, Rear), Right (Front, Mid, Rear)
// Oct:         1-8, depending on actual config (OctX8, OctFlatP, OctFlatX)
// Plane, Heli_120, Heli_90: undefined
// VTail4:      as QuadX

// first entry in layout struct represents the type: 0 - 2 values, 1 - 4 values, 2 - 6 values, 3 - 8 values
// following entries represent the associated motor/servo. (<20: motors, >=20: servos)

uint8_t esc_layout_0 [] = { 1,  2,  1, 30,  0, };            // Tri
uint8_t esc_layout_1 [] = { 1,  3,  0, 2,   1, };            // Q+, VTail
uint8_t esc_layout_2 [] = { 1,  3,  1, 2,   0, };            // Qx
uint8_t esc_layout_3 [] = { 1,  0,  1, 31, 32, };            // Bi

uint8_t esc_layout_4 [] = { 0, };                            // Gimbal - #0 only used once, hard wired
uint8_t esc_layout_5 [] = { 2,  2,  0,  1,  3,  5,  4, };    // Y6
uint8_t esc_layout_6 [] = { 2,  3,  4,  1,  2,  5,  0, };    // Hx
uint8_t esc_layout_7 [] = { 1, 20, 21, 22, 23, };            // FW

uint8_t esc_layout_8 [] = { 1,  3,  1,  0,  2, };            // Y4
uint8_t esc_layout_9 [] = { 2,  3,  5,  2,  1,  4,  0, };    // Hxx
uint8_t esc_layout_10[] = { 3, };                            // Oct, hard wired

uint8_t *esc_layouts [] = { esc_layout_0, esc_layout_1, esc_layout_2, esc_layout_3,
                            esc_layout_4, esc_layout_5, esc_layout_6, esc_layout_7,
                            esc_layout_8, esc_layout_9, esc_layout_10,
                          };

uint8_t esc_pos_4[] = { 3, 12, 3+16, 12+16, };               // LCD positions for 4 values
uint8_t esc_pos_6[] = { 2, 7, 12, 2+16, 7+16, 12+16, };      // LCD positions for 6 values

// ----------------

PROGMEM prog_char msg00[] = "1\002Config\007";
PROGMEM prog_char msg01[] = "2\002P-I-D";
PROGMEM prog_char msg02[] = "3\002Rates";
PROGMEM prog_char msg03[] = "4\002Sensors";
PROGMEM prog_char msg04[] = "5\002Receiver In";
PROGMEM prog_char msg05[] = "6\002ESC/Servo Out";
    
PROGMEM prog_char msg06[] = "Roll";
PROGMEM prog_char msg07[] = "Pitch";
PROGMEM prog_char msg08[] = "Yaw";
PROGMEM prog_char msg09[] = "D/R\003Expo\006";
PROGMEM prog_char msg10[] = "R/P\003Yaw\003Thro";
PROGMEM prog_char msg11[] = "Gyr";
PROGMEM prog_char msg12[] = "Acc";
PROGMEM prog_char msg13[] = "Altitude Heading";
    
PROGMEM prog_char msg14[] = "Roll\002Pitch\002Yaw";
PROGMEM prog_char msg15[] = "Thro\002Aux1\002Aux2";

PROGMEM prog_char msg16[] = "Lf:\006Ri:";
PROGMEM prog_char msg17[] = "Ru:\006Re:";
    
PROGMEM prog_char msg18[] = "Lf:\006Ri:";
PROGMEM prog_char msg19[] = "Fr:\006Re:";

PROGMEM prog_char msg20[] = "FL:\006FR:";
PROGMEM prog_char msg21[] = "RL:\006RR:";

PROGMEM prog_char msg22[] = "M1:\006M2:";
PROGMEM prog_char msg23[] = "S1:\006S2:";

PROGMEM prog_char msg24[] = "Roll.:";
PROGMEM prog_char msg25[] = "Pitch:";

PROGMEM prog_char msg26[] = "Up";
PROGMEM prog_char msg27[] = "Dn";

PROGMEM prog_char msg28[] = "Fr";
PROGMEM prog_char msg29[] = "Re";
    
PROGMEM prog_char msg30[] = "\003Rho\005Phi";
    
PROGMEM prog_char msg31[] = "7\002Profiles";
PROGMEM prog_char msg32[] = "8\002Settings";

PROGMEM prog_char msg33[] = "S1:\006S2:";
PROGMEM prog_char msg34[] = "S3:\006S4:";

PROGMEM prog_char msg35[] = "Aux3\002Aux4";
PROGMEM prog_char msg36[] = "Level";

PROGMEM prog_char msg37[] = "\0111-4 LMH";
PROGMEM prog_char msg38[] = "Roll+Pitch";
PROGMEM prog_char msg39[] = "Lf";
PROGMEM prog_char msg40[] = "Ri";
PROGMEM prog_char msg41[] = "Up:\006Dn:";
PROGMEM prog_char msg42[] = "Mag";
PROGMEM prog_char msg43[] = "Alt";
PROGMEM prog_char msg44[] = "Vel";
PROGMEM prog_char msg45[] = "Batt Power Alarm";
PROGMEM prog_char msg46[] = "GPS Dist Dir Sat";

PROGMEM prog_char msg47[] = "Calibrate Acc";
PROGMEM prog_char msg48[] = "Calibrate Mag";
PROGMEM prog_char msg49[] = "Activation Modes";
PROGMEM prog_char msg50[] = "Debug\002:\008";

PROGMEM prog_char msg51[] = "CycTime I2C Errs";
PROGMEM prog_char msg52[] = "GPS";

/* ----- */

// State messages
PROGMEM prog_char* msgs[] = {
 msg00, msg01, msg02, msg03, msg04, msg05, msg06, msg07, msg08, msg09, 
 msg10, msg11, msg12, msg13, msg14, msg15, msg16, msg17, msg18, msg19, 
 msg20, msg21, msg22, msg23, msg24, msg25, msg26, msg27, msg28, msg29, 
 msg30, msg31, msg32, msg33, msg34, msg35, msg36, msg37, msg38, msg39, 
 msg40, msg41, msg42, msg43, msg44, msg45, msg46, msg47, msg48, msg49,
 msg50, msg51, msg52, 
};

// State Transition Table, values above 80 are pseudo-states
PROGMEM prog_char nextState[NUM_STATES][NUM_KEYS + 1] = {
  // Right Up  Down Left Msg Status   --   Status Flag: >= 0: Refresh, <0: No Refresh, abs(n)>1: Msg2
  {    1,   8,   2, 100,   0,   0, }, //  0 - Config
  {    1,  45,  47,   0,   0,   0, }, //  1 - Acc, Bar, Mag

  {    9,   0,   3,   0,   1,  -1, }, //  2 - PID
  {   14,   2,   4,   0,   2,  -1, }, //  3 - Rates
  {   16,   3,   5,   0,   3,  -1, }, //  4 - Sensors
  {   21,   4,   6,   0,   4,  -1, }, //  5 - Rec In
  {   99,   5,   7,   0,   5,  -1, }, //  6 - ESc Out
  {   94,   6,   8,   0,  31,  -1, }, //  7 - Profile
  {   93,   7,   0,   0,  32,  -1, }, //  8 - Settings

  {   80,  39,  10,   2,  38,  -1, }, //  9 - PID Roll/Pitch
  {   81,   9,  11,   2,   6,  -1, }, // 10 - PID Roll
  {   82,  10,  12,   2,   7,  -1, }, // 11 - PID Pitch
  {   83,  11,  13,   2,   8,  -1, }, // 12 - PID Yaw
  {   84,  12,  37,   2,  43,  -1, }, // 13 - PID Alt

  {   90,  35,  15,   3,   9,  -1, }, // 14 - DR, Expo
  {   91,  14,  35,   3,  10,  -1, }, // 15 - R/P Rate, Yaw Rate, Throttle PID
 
  {  101,  46,  17,   4,  11,   0, }, // 16 - Gyro
  {  101,  16,  18,   4,  12,   0, }, // 17 - Acc
  {  101,  17,  41,   4,  42,   0, }, // 18 - Mag
  {  101,  20,  46,   4,  30,   0, }, // 19 - Altitde, Heading
  {  101,  41,  19,   4,  13,   0, }, // 20 - Angles

  {  101,  23,  22,   5,  14,   0, }, // 21 - R, P, Y
  {  101,  21,  23,   5,  15,   0, }, // 22 - Thr, Aux1, Aux2
  {  101,  22,  21,   5,  35,   0, }, // 23 - Aux3, Aux4

  {  101,  -1,  -1,   6,  16,  17, }, // 24 - Tri
  {  101,  -1,  -1,   6,  18,  19, }, // 25 - Quad+
  {  101,  -1,  -1,   6,  20,  21, }, // 26 - QuadX
  {  101,  -1,  -1,   6,  22,  23, }, // 27 - Bi
  {  101,  -1,  -1,   6,  24,  25, }, // 28 - Gimbal
  {  101,  -1,  -1,   6,  26,  27, }, // 29 - Y6
  {  101,  -1,  -1,   6,  28,  29, }, // 30 - Hex
  {  101,  -1,  -1,   6,  33,  34, }, // 31 - Flying Wing
  {  101,  -1,  -1,   6,  20,  41, }, // 32 - Y4
  {  101,  -1,  -1,   6,  39,  40, }, // 33 - Hex6X
  {  101,  -1,  -1,   6,  28,  29, }, // 34 - Oct

  {   92,  15,  14,   3,  45,   0, }, // 35 - Battery Level, PowerMeter [state no. referred in Serial]
  {   96,  97,  97,  44,  37,   0, }, // 36 - Enter Activation Modes

  {   85,  13,  38,   2,  44,  -1, }, // 37 - PID Vel
  {   86,  37,  39,   2,  52,  -1, }, // 38 - PID GPS
  {   87,  38,  40,   2,  36,  -1, }, // 39 - PID Level
  {   88,  39,   9,   2,  42,  -1, }, // 40 - PID Mag

  {  101,  18,  20,   4,  46,   0, }, // 41 - GPS Sensors

  {  102,  44,  43,   0,   0, -47, }, // 42 - Calibrate Acc
  {  103,  42,  45,   0,   0, -48, }, // 43 - Calibrate Mag
  {   36,  47,  42,   0,   0, -49, }, // 44 - Activation Modes
  {  104,  43,   1,   0,   0,  50, }, // 45 - Debug Values

  {  101,  19,  16,   4,  51,   0, }, // 46 - Cycle time, I2C Error Count
  {   47,   1,  44,   0,   0,   0, }, // 47 - GPS Modes

  // Right Up  Down Left Msg Status   --   Status Flag: >= 0: Refresh, <0: No Refresh, abs(n)>1: Msg2
};

// ----------------

int8_t   baudRates     [] = { 1, 2, 4, 8, 12, 16, 32, 48, 96, };
uint16_t auxMasks      [] = { 0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0100, 0x0200, 0x0400, 0x0800, 0x1000, 0x2000, }; 

// ----------------

//void checkFirstTime() {
//  uint8_t test_val; eeprom_read_block((void*)&test_val, (void*)(0), sizeof(test_val));
//  if (test_val == checkNewConf) return;
//  P8[ROLL] = 40; I8[ROLL] = 30; D8[ROLL] = 23;
//  P8[PITCH] = 40; I8[PITCH] = 30; D8[PITCH] = 23;
//  P8[YAW]  = 85; I8[YAW]  = 0;  D8[YAW]  = 0;
//  P8[PIDALT]   = 16; I8[PIDALT]   = 15;  D8[PIDALT]   = 7;
//  P8[PIDGPS]   = 50; I8[PIDGPS]   = 0;  D8[PIDGPS]   = 15;
//  P8[PIDVEL]   =  0; I8[PIDVEL]   = 0;  D8[PIDVEL]   = 0;
//  P8[PIDLEVEL] = 90; I8[PIDLEVEL] = 45; D8[PIDLEVEL] = 100;
//  P8[PIDMAG] = 40;
//  rcRate8 = 45; // = 0.9 in GUI
//  rcExpo8 = 65;
//  rollPitchRate = 0;
//  yawRate = 0;
//  dynThrPID = 0;
//  for(uint8_t i=0;i<CHECKBOXITEMS;i++) {activate1[i] = 0;activate2[i] = 0;}
//  accTrim[0] = 0; accTrim[1] = 0;
//  powerTrigger1 = 0;
//  #ifdef FLYING_WING
//    wing_left_mid  = WING_LEFT_MID; 
//    wing_right_mid = WING_RIGHT_MID; 
//  #endif
//  #ifdef TRI
//    tri_yaw_middle = TRI_YAW_MIDDLE; 
//  #endif
//  writeParams();
//}

uint8_t  defaultProfile[] = { 20,                     // v2.0
                              40, 30, 23,             // pid roll
                              40, 30, 23,             // pid pitch
                              85,  0,  0,             // pid yaw
                              16, 15,  7,             // pid alt
                              50,  0, 15,             // pid gps
                               0,  0,  0,             // pid vel
                              90, 45, 100,            // pid level
                              40,  0,  0,             // pid mag
                              45, 65,  0,  0,  0,     // dual rate, expo, roll/pitch rate, yaw rate, dyn phr pid
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // activation mode (11x2)
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                               0,  0,                 // power trigger (16 bit)
                            };

// ----------------

static char profileStr [] = "00: 01234567 ---";
static char statusMsg1 [] = "Nck- Baro- Mag- ";
static char statusMsg2 [] = "GPS Home- Hold- ";

static char lineBuffer [17];

#ifdef CHECK_INPUT
static char keySymbols [] = { ' ', 'R', 'U', 'D', 'L', 'S', };
#endif  

/* ----- Static Variables: ----- */

static uint16_t  ee_ptr                ;

static uint8_t*  bptr;
static uint8_t   buffer[ BUFFER_SIZE ] ;

static int8_t    state                 ; // current ui state
static boolean   graphMode             ; // graphical output of servo signals
static boolean   offline               ;
static int8_t    profile               ; // current profile index
static uint8_t   vers                  ; // MultiWii Software Version
static uint8_t   baudRate              ; // baud rate of serial connection to MWC board
static uint8_t   minMaxMode            ; // show act, min or max values
static uint8_t   keyPressedTime;       ; // time interval between key down and key up in 50ms steps
static uint8_t   auxState              ;
static uint8_t   delayTime             ; // x10 in milliseconds
static uint16_t  rawKey                ;
static uint16_t  lastReadTime          ;

// MWC data
static uint16_t  actMode[AUX_STATES]   ; // activation modes for level, baro, compas, camStab, camTrig, arming, gpsHome, gpsHold, passThru, headfree, beeper 
                                         // low byte: Aux1-2, high byte: Aux3-4, bits 0-2: Aux1 (Low, Mid, High), 3-5: Aux2, 8-10: Aux3, 11-13: Aux4, 15: rcOption flag (mode is currently on/off)

static uint8_t   accMode, baroMode, magMode, gpsModeHome, gpsModeHold, armed;
static uint8_t   accPres, baroPres, magPres, gpsPres;

static uint8_t   config                ; // Tri, Quad, Hex, etc.
static uint8_t   P8[8], I8[8], D8[8]   ; // Roll, Pitch, Yaw, Altitude, Velocity, GPS, Level, Magneto
static uint8_t   rcRate8               ; // dual rate
static uint8_t   rcExpo8               ; // expo
static uint8_t   rollPitchRate         ; // reduces "P" and "D" of Roll/Pitch depending on the associated stick deviation
static uint8_t   yawRate               ; // reduces "P" and "D" of Yaw depending on Rudder stick deviation
static uint8_t   dynThrPID             ; // reduces "P" and "D" of Roll/Pitch when Throttle stick is above 50%

static uint8_t   vbat                  ; // Battery Level
static uint16_t  powerMeterSum, powerTrigger;

static int16_t   accSmooth   [3] MM_DIM;  // projection of smoothed and normalized gravitation force vector on x/y/z axis, as measured by accelerometer       
static int16_t   gyroData    [3] MM_DIM;  // from WMP+
static int16_t   magADC      [3] MM_DIM;  // magnet sensor data
static int16_t   altitude        MM_DIM;  // from barometer
static int16_t   heading         MM_DIM;  // from compas
static int16_t   angle       [2] MM_DIM;  // absolute angle inclination in deg (when Nunchuk/Acc present)
static int16_t   servo       [8] MM_DIM;  // depends on configuration
static int16_t   motor       [8] MM_DIM;  // depends on configuration 
static int16_t   rcHysteresis[8] MM_DIM;  // receiver input

static int16_t   gpsDistToHome   MM_DIM; // MWC: uint16_t
static int16_t   gpsDirection    MM_DIM;
static uint8_t   gpsNumSat, gpsFix, gpsUpdate;

static uint16_t  cycleTime             ; // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop
static uint16_t  i2c_errorCount        ;

static int16_t   debugVals   [4]       ;  // multi-purpose debug values
static uint8_t   debugIndex            ;

// ----------------

void pinOut(uint8_t pin)                         {pinMode(pin, OUTPUT); }
void pinIn (uint8_t pin)                         {pinMode(pin, INPUT); digitalWrite(pin, HIGH);} // set pinMode and enable internal pullups

void delay50()                                   {delay(50);}
void delayT()                                    {delay(delayTime * 10);}

uint8_t getConfig()                              {return (config >= NUM_CONFIGS)? 0 : config;}
//#define getConfig() ((config >= NUM_CONFIGS)? 0 : config)

void initVars(int16_t arr[][3])                  {for (uint8_t i = 0; i < 8; i++) {arr[i][0] = arr[i][1] = arr[i][2] = 1500;}}
void initVars(int16_t arr[])                     {for (uint8_t i = 0; i < 8; i++) {arr[i] = 1500;}}

#ifdef MINMAX
void showMinMax()                                {lcdClear(); lcdPrintProg0(minMaxDsc + 4 * minMaxMode); blinkConfirm();}
#endif // MINMAX

// ----------------

void setup() {

  INIT_LCD
  INIT_BACKLIGHT
  INIT_JOYSTICK
  INIT_JS_SELECT
  INIT_LEDS
  INIT_TEST
  
  initVars(servo       );
  initVars(motor       );
  initVars(rcHysteresis);

#ifdef LCD_ETPP
    lcdPrintProg0(intro);     // Print it only once.  The loop below will cause an ETPP to flicker and appear very dim
    delay(2500);
#else
  for (uint8_t n = 160; n > 0; n--) {
    lcdPrintProg0(intro);
    delay(16);
    lcdClear();
  }
#endif

  restoreParams();            // restore Parameters from EEPROM (if present)
  serialBegin();              // initialize Serial communication

  state = -1;
  serialRead();
  
  state = 0;
  refreshDisplay();
}

/* ------ */

void loop() {
  int8_t statusFlag = pgm_read_byte_near(nextState[state] + 5); 
  if (statusFlag >= 0) {
    // update the current values from MultiWii board and refresh display
    uint16_t current = millis();
    if (current - lastReadTime > delayTime * 10) {
      lastReadTime = current;
      serialRead();
      showStatusLine();
    }
  }
  
  int8_t key = getKeyPressed();
  if (key >=0) {
    int8_t next = getNextState(state, key);
    if (next >= 80) { // handle pseudo-states
      if (next <= 88) {
        enterPid(next-80); 
      }
      else if (next <= 92) {
        enterRates(next-90); 
      }
      else if (next == 93) {
        int8_t state0 = 0;
        while (state0 >= 0) {
          if (key >= 0) {
            lcdClear();
            enterValue(-state0-1);
            lcdPrintProg0(setMsgs[state0]);
          }

          if (handleUpDownLeft(key, state0, NUM_SETTINGS)) {
            enterValue(state0+1); 
#ifdef BACKLIGHT
            setBacklight(backlight); 
#endif
          }
        }
      } 
      else if (next == 94) {
        int8_t state0 = 0;
        while (state0 >= 0) {
          if (key >= 0) {
            lcdClear();
            lcdPrintProg0(profStates[state0]);
            lcdWriteProg(profileMsg);
          }
          if (handleUpDownLeft(key, state0, 4)) {
            if (state0 == 3) {
              extractProfile(defaultProfile);
            }
            else {
              lcdWrite(':');
              uint8_t pv = 0;
              while (profile >= 0) {
                if (key >= 0) {
                  profileStr[0] = ('0' + profile/10);
                  profileStr[1] = ('0' + profile%10);
                  pv = getProfileName(profileStr+4);
                    
                  lcdPrint(16, profileStr);
                  if (pv > 0) {
                    format(13+16, pv, 3, 2);
                  }
                }
                if (handleUpDownLeft(key, profile, NUM_PROFILES)) {
                  if (state0 == 1) { 
                    // save profile
                    getProfileName(lineBuffer);
                    if (enterName(lineBuffer)) {
                      storeProfile(lineBuffer);
                    }
                  }
                  else if (pv > 0) {    // profile is valid
                    if (state0 == 0) {
                      extractCurrentProfile();
                      profile = -1;
                    }
                    else {
                      // clear profile
                      clearCurrentProfile();
                    }
                  }
                }
              }
              profile = 0;              // set to default profile
            }
          }
        }
      } 
      else if (next == 96) {
        lcdPrintProg0(auxStates[auxState]);
        lcdWrite(':');
        enterAuxState();
      }
      else if (next == 97) {
        int8_t offs = (key == KEY_DECR)? AUX_STATES-1 : 1;
        auxState    = (auxState + offs) % AUX_STATES;
      }
      else if (next == 99) { 
#ifdef TEST
        if (key == KEY_SELECT) {
          config++;
          if (config >= NUM_CONFIGS) {
            config = 1;
          }
          lcdPrintProg1(configIds + configOffsets[getConfig()]);
          blinkConfirm();
        }
#endif // TEST         
        int8_t conf = getConfig();         // select state depending on config
        state = 6;
        
        if (conf > 0) {
          if (conf < 11) { 
            state = conf + CONF_OFFSET;
          } 
          else if (conf < 14) {
            state = 34; // OctX, OctF+, OctFx
          }
          else if (conf == 17) {
            state = 26; // VTail4
          }
        }
      }
      else if (next == 100) {
#ifdef CHECK_INPUT
        if (keyPressedTime > 10) {
          lcdClear();
          lcdPrintProg0(jsStatusMsg);

          while (true) {
            int8_t key = getKey();
            
            lcdWrite    (     16, keySymbols[key+1]);
  #ifdef KEY_PIN            
            format      ( 2 + 16, rawKey);
  #endif

  #ifdef JOYSTICK            
            formatSigned( 7 + 16, jsHorAct  - jsHorCtr);
            formatSigned(12 + 16, jsVertAct - jsVertCtr);
  #endif
  
  #ifdef STATUS_LEDS
            key++;
            red   (key & 0x1); // output the key id in binary
            yellow(key & 0x2);
            green (key & 0x4);
  #endif
            delay50();
          }
        }
        else
#endif // CHECK_INPUT
        {
          lcdPrintProg1(emptyLine);
          lcdPrintProg1(metaData);
          lcdWrite(baudRate + '0');
#ifdef CREDITS
          enterKey();
          lcdPrintProg1(credits);
          enterKey();
#endif // CREDITS     
        }
      } 
      else if (next == 101) {
#ifdef MINMAX
  #ifdef GRAPHIC
        if (key == KEY_SELECT || keyPressedTime > 10) {
          graphMode = !graphMode;
        }
        else
  #endif
        {
          minMaxMode = (minMaxMode + 1) % 3; 
          showMinMax();
        }
#else        
  #ifdef GRAPHIC
        graphMode = !graphMode;
  #endif
#endif //MINMAX
      }
      else if (next == 102 || next == 103) {
        serialWrite((next == 102)? 'S' : 'E');
        next = 4;
        lcdWrite(15+16, '*');
        blinkLed(RED_LED, 200);
      }
      else if (next == 104) {
        debugIndex++;
        if (debugIndex == 4) {
          debugIndex = 0;
        }
      }
    
      next = state;
    }

    if (next >= 0) {
      state = next;
      refreshDisplay();
    }
  }
}

boolean handleUpDownLeft(int8_t& key, int8_t& state, int8_t modulo) {
  key = getKeyPressed();
  if      (key == KEY_LEFT) { state = -1; }
  else if (key == KEY_DOWN) { state++; if (state >= modulo) state = 0;        } 
  else if (key == KEY_UP  ) { state--; if (state < 0)       state = modulo-1; }
  else if (key >= 0       ) { return true; }
  return false;
}

void refreshDisplay() {
  lcdClear();

  if (state >= 0) {
    // two levels of indirection...
    int8_t           index1 = pgm_read_byte_near(nextState[state] + 4);             // this is the index in the msgs array
    prog_char const* cptr0  = (prog_char const*) pgm_read_word_near(msgs + index1); // this is the adress of the msg
    int8_t           index2 = pgm_read_byte_near(nextState[state] + 5);
    if (index2 < 0) {index2 = -index2;} 
    prog_char const* cptr1 = (index2 > 1)? (prog_char const*) pgm_read_word_near(msgs + index2) : emptyLine;
  
    lcdPrintProg2(cptr0, cptr1);
    showStatusLine();
  }
}

char getStatusChar(int8_t pres, int8_t mode) {
  return (pres)? (mode? '1' : '0') : '-';
}

void showStatusLine() {
  if (state == 0) {
    lcdPrintProg1(configIds + configOffsets[getConfig()]);
    
    if (offline) {
      lcdPrintProg(9 + 16, offlineMsg);
    }
    else {
      lcdWrite( 9 + 16, 'v');      
      format  (10 + 16, vers, 3, 2);
      lcdWrite(15 + 16, (armed)? '@' : ' '); // "armed"
    }
  }
  else if (state == 1) {
    char c0, c2;
    if (accPres > 1) {
      c0 = 'A';  // Acc
      c2 = 'c';
    }
    else {
      c0 = 'N'; // Nck - Nunchuk 
      c2 = 'k';
    }

    statusMsg1[ 0] = c0;
    statusMsg1[ 2] = c2;
    statusMsg1[ 3] = getStatusChar(accPres , accMode );
    statusMsg1[ 9] = getStatusChar(baroPres, baroMode);
    statusMsg1[14] = getStatusChar(magPres , magMode );
    lcdPrint(16, statusMsg1);
  }
  else if ((state >= 9 && state <= 13) || (state >= 37 && state <= 40)) {
    int8_t idx = (state == 9)? 0 : (state <= 13)? state - 10 : state - 33;

    format( 0 + 16, P8[idx], 4, 2);
    if (state != 40) {
      format( 5 + 16,  I8[idx], 5, 4);
      if (vers >= 20 || state != 39) {
        formatSigned(11 + 16, -D8[idx]);
      }
    }
  }
  else if (state == 14) { 
    // Dual Rate, Expo
    format( 0 + 16, rcRate8 * 2);
    format( 6 + 16, rcExpo8);
  }
  else if (state == 15) { 
    // Roll/Pitch Rate, Yaw Rate, Throttle PID
    format( 0 + 16, rollPitchRate);
    format( 6 + 16, yawRate);
    format(12 + 16, dynThrPID);
  }
  else if (state >= 16 && state <= 18) {  // Gyro, Acc, Mag
    lcdPrintProg(3, rollPitchYaw);
    for (uint8_t i = 0; i < 3; i++) {
      int16_t val;
      if      (state == 16) { val = gyroData [i] MM_VAL; }
      else if (state == 17) { val = accSmooth[i] MM_VAL; }
      else                  { val = magADC   [i] MM_VAL; }
      formatSigned(i*6 + 16, val); 
    }
  }
  else if (state == 19) { // Angles
    drawAngles();
  }
  else if (state == 20) { // Alt, Head
    formatSigned( 1 + 16, altitude MM_VAL, 6, 0);
    formatSigned(10 + 16, heading  MM_VAL);
  }
  else if (state >= 21 && state <= 23) { // receiver input
    int8_t iOff = (state-21) * 3;
    int8_t sOff = 16;
    int8_t maxV = 3;

    if (state == 23) {
      maxV = 2;
    }
    for (int8_t i = 0; i < maxV; i++) {formatScaled(i*6 + sOff, rcHysteresis[i + iOff] MM_VAL, !(i == 0 && state == 22));}
  }
  else if (state >= 24 && state < 35) {
    printEscLayout(esc_layouts[state-24]);
  }
  else if (state == 35) { // Battery Level
    format(     16, vbat,          4, 2);
    format( 5 + 16, powerMeterSum, 5, 0);
    format(11 + 16, powerTrigger,  5, 0);
    
  #ifdef STATUS_LEDS
    if (vbat > 0) {
      green (vbat >  VBATLEVEL1);      
      yellow(vbat >  VBATLEVEL3 && vbat <= VBATLEVEL0);
      red   (vbat <= VBATLEVEL2);
    }
  #endif
  }
  else if (state == 36) { // Activation Mode
    lcdPrintProg0(auxStates[auxState]);
    printAuxState();
  }
  else if (state == 41) {
    formatSigned( 0 + 16, gpsDistToHome MM_VAL, 6, 0);
    formatSigned( 7 + 16, gpsDirection  MM_VAL);
    format      (12 + 16, gpsNumSat);
  }
  else if (state == 45) {
    lcdWrite    ( 6 + 16, '0' + debugIndex);
    formatSigned(10 + 16, debugVals[debugIndex], 6, 0);
  }
  else if (state == 46) {      // cylce time, i2c error count
    format      ( 2 + 16, cycleTime);
    format      ( 9 + 16, i2c_errorCount, 5, 0);
  }
  else if (state == 47) {
    statusMsg2[ 8] = getStatusChar(accPres , gpsModeHome);
    statusMsg2[14] = getStatusChar(baroPres, gpsModeHold);
    lcdPrint(16, statusMsg2);
  }
}

void printEscLayout(uint8_t layout[]) {
  uint8_t tag = layout[0];
  
  if (tag == 0) {
    printEsc(     7, 20);
    printEsc(16 + 7, 21);
  }
  else if (tag == 1) {
    for (uint8_t i = 0; i < 4; i++) { printEsc(esc_pos_4[i], layout[i+1]); }
  }
  else if (tag == 2) {
    for (uint8_t i = 0; i < 6; i++) { printEsc(esc_pos_6[i], layout[i+1]); }
  }
  else {
    for (uint8_t i = 0; i < 8; i++) { printEsc(i*4, i); }
  }
}

int8_t getNextState(int8_t state, int8_t key) {
  int8_t col = (key == KEY_SELECT)? KEY_RIGHT : (key == STATUS)? STATUS-1 : key;

#ifdef MINMAX
  if (state >= 4 && state <= 6 && col == KEY_RIGHT) {
    showMinMax();
  }
#endif  

#ifdef STATUS_LEDS
  if (state == 35) {
    green(false); yellow(false); red(false);
  }
#endif

  return pgm_read_byte_near(nextState[state] + col);
}
