/*
 * ----- Persistence: ----- 
 */

uint8_t readEEPROM ()                       {return EEPROM.read(ee_ptr++);}
void    writeEEPROM(uint8_t val)            {if (EEPROM.read(ee_ptr) != val) {EEPROM.write(ee_ptr, val);} ee_ptr++;}

int getProfileSlot(int8_t slot) {
  return slot * 84 + 8;
}

void clearCurrentProfile() {
  ee_ptr = getProfileSlot(profile);
  for (uint8_t i = 0; i < 9; i++) {         // just delete name and version
    writeEEPROM(0);
  }
}

boolean extractCurrentProfile() {
  ee_ptr = getProfileSlot(profile) + 8;     // skip name

  uint8_t* lptr = buffer;
  for (uint8_t i = 0; i < 84; i++) {
    *lptr++ = readEEPROM();
  }
  return extractProfile(buffer);
}
 
boolean extractProfile(uint8_t* lptr) { 
  uint8_t tag = *lptr++;                    // 00: version
  if (tag < 18 || tag > 20) {
    flashRed();
    return false;                           // wrong version
  }

  int8_t i;
  for (i = 0; i < 7; i++) {                 // 01-21: PID (roll, pitch, yaw, alt, vel, gps, level)
    P8[i] = *lptr++; 
    I8[i] = *lptr++; 
    D8[i] = *lptr++;
  }
  P8[7] = *lptr++;                          // 22: P mag
  I8[7] = 0;
  D8[7] = 0;

  rcRate8       = *lptr++;                  // 23: rcRate (dual rate)
  rcExpo8       = *lptr++;                  // 24: rcExpo
  rollPitchRate = *lptr++;                  // 25: rollPitchRate
  yawRate       = *lptr++;                  // 26: yawRate
  dynThrPID     = *lptr++;                  // 27: dyn. Throttle PID

  for (i = 0; i < AUX_STATES; i++) {
    uint8_t low  = *lptr++;
    uint8_t high = *lptr++;
    actMode[i]   = high * 256 + low;        // 28-49: activation modes
  }
  uint8_t low  = *lptr++;
  uint8_t high = *lptr++;
  powerTrigger = high * 256 + low;          // 50-51: powerTrigger
  
  serialWrite();
  return true;
}

void storeProfile(char* name) {
  uint8_t* lptr = buffer;

  int8_t i = 0;
  while (name[i] != '\0' && i < 8) { *lptr++ = name[i++]; }
  while (i < 8)                    { *lptr++ = ' ';       }
  
  *lptr++ = vers;                         // 00: version
  for (i = 0; i < 7; i++) {               // 01-21: PID (roll, pitch, yaw, alt, vel, gps, level)
    *lptr++ = P8[i];
    *lptr++ = I8[i]; 
    *lptr++ = D8[i];
  }
  *lptr++ = P8[7];                        // 22: P mag

  *lptr++ = rcRate8;                      // 23: rcRate (dual rate)
  *lptr++ = rcExpo8;                      // 24: rcExpo
  *lptr++ = rollPitchRate;                // 25: rollPitchRate
  *lptr++ = yawRate;                      // 26: yawRate
  *lptr++ = dynThrPID;                    // 27: dyn. Throttle PID

  for (i = 0; i < AUX_STATES; i++) {      // 28-49: activation modes
    *lptr++ = actMode[i] % 256;
    *lptr++ = actMode[i] / 256;
  }
  *lptr++ = powerTrigger % 256;
  *lptr++ = powerTrigger / 256;           // 50-51: powerTrigger
  
  int len = (lptr - buffer);
  lptr    = buffer;
  ee_ptr  = getProfileSlot(profile);
  for (i = 0; i < len; i++) {
    writeEEPROM(*lptr++);
  }
  
  flashYellow();
}

uint8_t getProfileName(char* name) {
  ee_ptr = getProfileSlot(profile);
  for (int8_t i = 0; i < 8; i++) { name[i] = readEEPROM(); }

  uint8_t tag = readEEPROM();
  if (tag < 18 || tag > 20) {
    for (int8_t i = 0; i < 8; i++) { name[i] = '-'; }
    tag = 0;
    flashRed();
  }
  return tag;
} 

// --------------------------------

void restoreParams() {
  ee_ptr   = 0;
  baudRate = DEFAULT_BAUDRATE;  // default value: 8 = 115200 Baud

  byte magic0 = readEEPROM();   // check validity of data by storing some magic numbers as a prefix
  byte magic1 = readEEPROM();
  ee_ptr++;                     // was: MWC version

  if (magic0 == 47 && magic1 == 11) {
    ee_ptr++;                   // was: logMode
    
    delayTime = readEEPROM();

#ifdef BACKLIGHT
    backlight = 255 - readEEPROM();
#else
    ee_ptr++;
#endif

#ifndef USE_DEFAULT_BAUDRATE    
    int8_t tmp1 = -1 + readEEPROM();
    if (tmp1 < 8) {
      baudRate = tmp1;
    }
#endif    
  }
}

void storeParams() {
  ee_ptr = 0;
  writeEEPROM(47);  
  writeEEPROM(11);
  ee_ptr++;                     // was: MWC version
  ee_ptr++;                     // was: logMode
  writeEEPROM(delayTime);
#ifdef BACKLIGHT
  writeEEPROM(255 - backlight);
#else
  ee_ptr++;
#endif
  writeEEPROM(baudRate + 1);
}

